#pragma once
#include "fbohelp.h"
#include <vector>
using namespace std;

class GraphicProcessor
{
public:
	GraphicProcessor(size_t vecsz){
		// initialize texture format;
		texTarget_				= GL_TEXTURE_RECTANGLE_ARB;
		texInternalFormat_	= GL_RGBA32F_ARB;
		texFormat_				= GL_RGBA;

		//Initialize Windows and GLUT
		int argc;
		char* argv = "\0";
		glutInit ( &argc, &argv );
		glutWindowHandle_ = glutCreateWindow("GPGPU");

		//Initialize GLEW
		int err = glewInit();
		if (GLEW_OK != err) {
			printf((char*)glewGetErrorString(err));
			//		PAUSE();
			exit(ERROR_GLEW);
		} 

		if (texFormat_ == GL_RGBA)
			texSize_ = (int)sqrt(vecsz/4.0);
		else
			texSize_ = (int)sqrt((double)vecsz);

		// initialize off-screen frame buffer(FBO)
		glGenFramebuffersEXT(1, &fb_); 
		// bind off-screen frame buffer (that is, skip the window-specific render target)
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb_);

		// view port for 1:1 pixel=texture mapping
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluOrtho2D(0.0, GLdouble(texSize_), 0.0, GLdouble(texSize_));
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glViewport(0, 0, texSize_, texSize_);
	}

	~GraphicProcessor(){
		glDeleteFramebuffersEXT(1,&fb_);
		glutDestroyWindow (glutWindowHandle_);
	}

	void RegisterKernel(Kernel* pK){
		pKernel_ = pK;
	}

	vector<GLuint> AllocMems(size_t nMemBlocks){
		vector<GLuint> glTexIDs(nMemBlocks);

		glGenTextures (nMemBlocks, &(glTexIDs[0]));
		for(int i = 0; i < nMemBlocks; ++i) { SetupTexture (glTexIDs[i]); }
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		checkGLErrors ("createFBOandTextures()");
		
		return glTexIDs;
	}

	void FreeMems(const vector<GLuint>& memIDs)
	{
		glDeleteTextures(memIDs.size(), &(memIDs[0]));
	}

	GLuint AllocateMem()
	{
		GLuint glTexID;

		glGenTextures(1, &glTexID);
		SetupTexture(glTexID);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		checkGLErrors ("createFBOandTextures()");

		return glTexID;
	}

	void FreeMem(GLuint memID)
	{
		glDeleteTextures(1, &memID);
	}

	void WriteMem(float* data, GLuint memID)
	{
		// version (a): HW-accelerated on NVIDIA 
		// glBindTexture(tex_param.texTarget, texID);
		// glTexSubImage2D(tex_param.texTarget,0,0,0,texSize_,texSize_,tex_param.texFormat,GL_FLOAT,data);
		// version (b): HW-accelerated on ATI 
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, texTarget_, memID, 0);
		glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
		glRasterPos2i(0,0);
		glDrawPixels(texSize_,texSize_,texFormat_,GL_FLOAT,data);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, texTarget_, 0, 0);
	}

	void ReadMem(float* data, GLuint memID)
	{
		glBindTexture(texTarget_,memID);
		glGetTexImage(texTarget_,0,texFormat_,GL_FLOAT,data);
	}

	void ReadBuffer(float* data){
		glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
		glReadPixels(0, 0, texSize_, texSize_,texFormat_,GL_FLOAT,data);
	}

	GLuint RegisterMemBlock(GLuint memID){
		static GLenum texStage = GL_TEXTURE0;
		glActiveTexture(texStage);	
		glBindTexture(texTarget_, memID);
		return texStage++;
	}
	
	GLuint RegisterMemBlock(GLuint memID, GLuint texStage)
	{
		glActiveTexture(texStage);	
		glBindTexture(texTarget_, memID);
		return texStage;
	}

	void	PerformComputation(GLuint destMemID){
		//	glEnable(tex_param.texTarget);
		// attach two textures to FBO
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, texTarget_, destMemID, 0);
		// check if that worked
		if (!checkFramebufferStatus()) {
			printf("glFramebufferTexture2DEXT():\t [FAIL]\n");
			exit (ERROR_FBOTEXTURE);
		}

		pKernel_->CreateKernelEnv();

		// Calling glFinish() is only necessary to get accurate timings,
		// and we need a high number of iterations to avoid timing noise.
		glFinish();
		// set render destination
		glDrawBuffer (GL_COLOR_ATTACHMENT0_EXT);
		// make quad filled to hit every pixel/texel 
		// (should be default but we never know)
		glPolygonMode(GL_FRONT,GL_FILL);
		// and render the quad
		if (texTarget_ == GL_TEXTURE_2D) {
			// render with normalized texcoords
			glBegin(GL_QUADS);
			glTexCoord2f(0.0, 0.0); 
			glVertex2f(0.0, 0.0);
			glTexCoord2f(1.0, 0.0); 
			glVertex2f(texSize_, 0.0);
			glTexCoord2f(1.0, 1.0); 
			glVertex2f(texSize_, texSize_);
			glTexCoord2f(0.0, 1.0); 
			glVertex2f(0.0, texSize_);
			glEnd();
		} else {
			// render with unnormalized texcoords
			glBegin(GL_QUADS);
			glTexCoord2f(0.0, 0.0); 
			glVertex2f(0.0, 0.0);
			glTexCoord2f(texSize_, 0.0); 
			glVertex2f(texSize_, 0.0);
			glTexCoord2f(texSize_, texSize_); 
			glVertex2f(texSize_, texSize_);
			glTexCoord2f(0.0, texSize_); 
			glVertex2f(0.0, texSize_);
			glEnd();
		}
		// done, stop timer, calc MFLOP/s if neccessary
		glFinish();

		pKernel_->DestoryKernelEnv();
		// calc mflops 
		//double mflops = (2.0*N*numIterations) / (total * 1000000.0);
		//printf("MFLOP/s for N=%d:\t\t%f %f %d %d\n", N, mflops, total, end, CLOCKS_PER_SEC);
		// done, just do some checks if everything went smoothly.
		checkFramebufferStatus();
		checkGLErrors("render()");
		//	glDisable(tex_param.texTarget);
	}

protected:

	/**
	* Sets up a floating point texture with NEAREST filtering.
	* (mipmaps etc. are unsupported for floating point textures)
	*/
	void SetupTexture (const GLuint texID) {
		// make active and bind
		glBindTexture(texTarget_, texID);
		// turn off filtering and wrap modes
		glTexParameteri(texTarget_, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(texTarget_, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(texTarget_, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(texTarget_, GL_TEXTURE_WRAP_T, GL_CLAMP);
		// define texture with floating point format
		glTexImage2D(texTarget_,0,texInternalFormat_,texSize_,texSize_,0,texFormat_,GL_FLOAT,0);
		// check if that worked
		if (glGetError() != GL_NO_ERROR) {
			printf("glTexImage2D():\t\t\t [FAIL]\n");
			exit (ERROR_TEXTURE);
		}
		//	printf("Created a %i by %i floating point texture.\n",texSize_,texSize_);
	}

	int			texSize_;				// size of texture. a texture is texSize_*texSize_.
	int			vectorSize_;			// size of data.
	Kernel*	pKernel_;				// the kernel function gpu used.

	GLenum	texTarget_;			// target texture coordinate format.
	GLenum texInternalFormat_;	// internal format of texture.
	GLenum texFormat_;			// texture format.
	GLuint	glutWindowHandle_;
	GLuint	fb_;
};